(defpackage :aoc/2018/07
  (:use :cl :aoc :alexandria :queues)
  (:export
   #:sample-data
   #:part1
   #:step-time
   ))
(in-package :aoc/2018/07)

(defparameter sample-input
  (sb-unicode:lines "Step C must be finished before step A can begin.
Step C must be finished before step F can begin.
Step A must be finished before step B can begin.
Step A must be finished before step D can begin.
Step B must be finished before step E can begin.
Step D must be finished before step E can begin.
Step F must be finished before step E can begin."))

(defun parse-line (s)
  (let ((words (sb-unicode:words s)))
    (list
     (read-from-string (nth 2 words))
     (read-from-string (nth 14 words)))))

(defvar sample-data '())
(setf sample-data (map 'list #'parse-line sample-input))

(defvar input-data '())
(setf input-data (map 'list #'parse-line (aoc:test-input 2018 7)))

(defun prerequisites-availablep (available target dependencies)
  (every (lambda (s) (member s available))
	 (loop for dep in dependencies
	       if (eq target (second dep))
		 collect (first dep))))

(defun steps-without-dependencies (dependencies)
  (let ((produced (loop for dep in dependencies collect (second dep))))
    (remove-duplicates 
     (loop for dep in dependencies
	   for p = (first dep)
	   if (not (member p produced))
	     collect p))))

(defun available-steps (done dependencies)
  (remove-if
   (lambda (x) (member x done))
   (append
    (steps-without-dependencies dependencies)
    (loop for dep in dependencies
	  if (and
	      (member (first dep) done)
	      (prerequisites-availablep done (second dep) dependencies))
	    collect (second dep)))))

(defun next-step (done dependencies)
  (let ((available (available-steps done dependencies)))
    (first (sort available #'string-lessp))))

(defun assembly-order (dependencies &optional (done '()))
  (let ((next (next-step done dependencies)))
    (if (not next)
	(reverse done)
	(assembly-order dependencies (push next done)))))

(defun part1 (&optional (data input-data))
  (format nil "~{~A~}" (assembly-order data)))

;;; -------------------------------------------------------------------

(defun step-time (step &optional (offset 60))
  (+ offset (mod (char-code (character step)) 32)))

(defun available-timed-steps (time done dependencies &optional (offset 60))
  (map 'list
       (lambda (s) (list (+ time (step-time s offset)) s))
       (available-steps (map 'list #'second done) dependencies)))

(defun timed-step-lessp (a b)
  (cond
    ((eq (first a) (first b)) (string-lessp (second a) (second b)))
    (t (< (first a) (first b)))))

(defun merge-timed-steps (steps add workers)
  (let ((new-steps steps))
    (loop for ev in (sort add #'timed-step-lessp)
	  while (< (length new-steps) workers)
	  if (not (member (second ev) (mapcar #'second new-steps)))
	    do (setf new-steps (push ev new-steps)))
    (sort new-steps #'timed-step-lessp)))

(defun assembly-order-timed (dependencies &key (time 0) (done '()) (todo '()) (offset 60) (workers 5))

  ;; done is a sorted list of (ts step) pairs
  
  (format t "todo: ~A with ~A at ~A~%" todo done time)
  (cond
    ((and (not done) (not todo))
     (assembly-order-timed dependencies
			   :time time
			   :done done
			   :todo (available-timed-steps time done dependencies offset)
			   :offset offset
			   :workers workers))
    ((and done (not todo))
     (progn
       (format t "we're done : ~A~%" done)
       done))
    (t
     (let* ((next-done (append done (remove-if (lambda (x) (< time (first x))) todo)))
	    (next-todo (merge-timed-steps
			(remove-if (lambda (x) (>= time (first x))) todo)
			(available-timed-steps time next-done dependencies offset)
			workers)))
       (format t "next todo: ~A, done: ~A ~%" next-todo next-done)
       (let* ((next-step (first next-todo))
	      (nt (first next-step)))
	 (format t "~A~%" next-todo)
	 (assembly-order-timed
	  dependencies
	  :time nt
	  :done next-done
	  :todo next-todo
	  :offset offset
	  :workers workers))))))


(defun part2 ()
  (let ((finish (first (first (last (assembly-order-timed input-data))))))
    (format nil "~A" finish)))


