(defpackage :aoc/2024/19
  (:use :cl :aoc :alexandria :trivia :lla)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   #:towel-tree
   #:make-onsen
   #:onsen-towels
   #:onsen-patterns
   #:remaining-patterns
   #:reset-patterns
   #:possible-patterns
   ))

(in-package :aoc/2024/19)


(defun parse-line (line)
  (ppcre:split ", " line))

(defstruct onsen towels patterns)
(defun parse-input (lines)
  (make-onsen
   :towels (parse-line (first lines))
   :patterns (mapcar (compose #'first #'parse-line) (subseq lines 2))))

(defparameter input-text (test-input 2024 19))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "r, wr, b, g, bwu, rb, gb, br

brwrr
bggr
gbbr
rrbgbr
ubwu
bwurrg
brgr
bbrgwb"))
(defparameter sample-data 
  (parse-input sample-text))

(defvar *pattern-cache* (make-hash-table :test #'equal))
(defvar *possible-cache* (make-hash-table :test #'equal))

(defun reset-patterns ()
  (setf *pattern-cache* (make-hash-table :test #'equal))
  (setf *possible-cache* (make-hash-table :test #'equal)))

(defun string-prefix-p (prefix string)
  (and
   (>= (length string) (length prefix))
   (equal prefix (subseq string 0 (length prefix)))))

(defun remaining-patterns (towels pattern)
  (or
   (gethash pattern *pattern-cache*)
   (let ((remaining (loop
		      for towel in towels
		      if (string-prefix-p towel pattern)
			collect (subseq pattern (length towel))
		      )))
     (setf (gethash pattern *pattern-cache*) remaining))))

(defun possible-patterns (towels pattern)
  (multiple-value-bind (n found)
      (gethash pattern *possible-cache*)
    (if found
	n
	(setf (gethash pattern *possible-cache*)
	      (if (emptyp pattern)
		  1		  
		  (if-let (ps (remaining-patterns towels pattern))
		    (loop
		      for p in ps
		      for n = (possible-patterns towels p)
		      sum n 
		      )
		    0))))
    ))

(defun possible-patterns-per-towel (onsen)
  (loop for p in (onsen-patterns onsen)
	for n = (possible-patterns (onsen-towels onsen) p)
	collect	n))

(defun part1 (onsen)
  (format nil "~A" (count-if-not #'zerop (possible-patterns-per-towel onsen))))

(defun part2 (onsen)
  (format nil "~A" (reduce #'+ (possible-patterns-per-towel onsen))))

(defun solve-day ()
  (reset-patterns)
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 19 p1))
    (if p2 (submit-part2 2024 19 p2))))
