
(defpackage :aoc/2024/20
  (:use :cl :aoc :aoc/maze :alexandria :trivia :lla)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   ))

(in-package :aoc/2024/20)


(defparameter input-text (test-input 2024 20))
(defparameter input-data (parse-maze input-text))

(defparameter sample-text (aoc:split-lines "###############
#...#...#.....#
#.#.#.#.#.###.#
#S#...#.#.#...#
#######.#.#.###
#######.#.#...#
#######.#.###.#
###..E#...#...#
###.#######.###
#...###...#...#
#.#####.#.###.#
#.#...#.#.#...#
#.#.#.#.#.#.###
#...#...#...###
###############
"))
(defparameter sample-data 
  (parse-maze sample-text))


(defun next-moves (map pos)
  "return conses of next pos and cost"
  (loop for d in '(#\^ #\v #\> #\<)
	for p = (pos-move pos d)
	if (and
	    (array-in-bounds-p map (pos-x p) (pos-y p))
	    (not (eq #\# (aref map (pos-y p) (pos-x p)))))
	  collect (cons p 1)))

(defun best-path (map)
  (let* ((start (find-tile map #\S))
	 (finish (find-tile map #\E))
	 (pathfinder (make-a-star #'next-moves
				  #'manhattan-distance))
	 )
    (funcall pathfinder map start finish)))


(defun find-shortcuts (map &optional (max-distance 2))
  (let ((path (best-path map)))
    (loop
      for s1 in path
      append (loop
	       for s2 in path
	       for dist = (manhattan-distance (cdr s1) (cdr s2)) 
	       for saving = (- (car s2) (car s1) dist)
	       if (and
		   (plusp saving)
		   (<= dist max-distance))
		 collect (list saving (cdr s1) (cdr s2))))))

(defun count-sc-groups (shortcuts &optional (min-savings))
  ;; present results to make it easy to compare to challenge
  (let ((count-map (make-hash-table)))
    (loop
      for s in shortcuts
      do (incf (gethash (first s) count-map 0)))
    (sort
     (loop
       for k being the hash-keys in count-map
       for v being the hash-values in count-map
       if (>= k min-savings)       
	 collect (cons k v)
       )
     #'(lambda (a b) (< (car a) (car b)))
     )))


(defun part1 (map)
  (loop for sc in (find-shortcuts map) count (>= (car sc) 100) ))

(defun part2 (map)
  (loop for sc in (find-shortcuts map 20) count (>= (car sc) 100) ))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 20 p1))
    (if p2 (submit-part2 2024 20 p2))))
