
(defpackage :aoc/2024/22
  (:use :cl :aoc :alexandria :lla :arrow-macros)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   #:next-random
   #:repeat-random
   ))

(in-package :aoc/2024/22)


(defun parse-line (line)
  (parse-integer line))


(defun parse-input (lines)
  (mapcar #'parse-line lines))

(defparameter input-text (test-input 2024 22))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "1
10
100
2024
"))

(defparameter sample-data 
  (parse-input sample-text))

(defparameter PRUNE (ash 1 24))
(defun rnd-step (n d)
  (mod (logxor n (ash n d)) PRUNE))

(defun next-random (n)
  (-> n
    (rnd-step 6)
    (rnd-step -5)
    (rnd-step 11)))

(defun repeat-random (seed n)
  (loop
    for i from 1 to n
    do (setf seed (next-random seed))
    finally (return seed)))

(defun part1 (data)
  (loop
    for buyer in data
    sum (repeat-random buyer 2000)))

(defun day-prices (seed)
  (make-array '(2001)
	      :initial-contents
	      (loop
		for i from 1 to 2001
		collect (mod seed 10)
		do (setf seed (next-random seed))
		)
	      
	      ))

(defun sum-table (a b)
  (loop
    for i from 0 below (array-total-size a)
    for x = (aref b i)
    if (plusp x)
      do (incf (aref a i) x))
  as)

(defun make-diff (prices)
  (let ((diffs (copy-array prices)))
    (loop
      for i from 1 below (array-dimension prices 0)
      do (decf (aref diffs i) (aref prices (1- i))))
    diffsG))

(defun make-buyer-table (seed)
  (let* ((prices (day-prices seed))
	 (diffs (make-diff prices))
	 (buyer-table (make-array (list 19 19 19 19) :initial-element -1)))
    
    (loop
      for i from 4 below (array-dimension prices 0)
      for p = (aref prices i) 
      for d1 = (+ 9 (aref diffs i))
      for d2 = (+ 9 (aref diffs (1- i)))
      for d3 = (+ 9 (aref diffs (- i 2)))
      for d4 = (+ 9 (aref diffs (- i 3)))
      for bp = (aref buyer-table d1 d2 d3 d4)
      if (minusp bp)
	do (setf (aref buyer-table d1 d2 d3 d4) p))
    buyer-table))

(defun price-diff-table (seeds)
  (let ((table (make-array (list 19 19 19 19) :initial-element 0)))
    (loop
      for buyer from 0
      for s in seeds
      do (sum-table table (make-buyer-table s)))
    table))

(defun maxtable (table)
  (loop
    for i from 0 below (array-total-size table)
    maximize (row-major-aref table i)))




(defparameter sample-data2 '(1 2 3 2024))

(defun part2 (data)
  (maxtable (price-diff-table data)))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 22 p1))
    (if p2 (submit-part2 2024 22 p2))))
